# Import classes and functions from RPE_Utils
from AutoRPE.UtilsRPE.MainFunctions import parse_sources, preprocess_sources
from AutoRPE.UtilsRPE.Functions import load_vault, in_which_block
from AutoRPE.UtilsRPE.track_dependecy import track_dependencies, propagate_dependencies
from os import listdir, mkdir
from os.path import join, isdir
from shutil import rmtree
import re


# Function to find the actual precision of a variable in the source file
def find_precision(variable, source_path, vault, lookup_table):
    filename = variable.routine.module.name + ".f90"
    var_name = var.name
    filepath = join(source_path, filename)
    pattern = r"real *\(([^\)]*)\).*::.*\b%s\b" % var_name
    cases = ""
    with open(filepath, "r") as f:
        lines = [l for l in f]
        current_block = variable.routine.module.main()
        condition1 = False
        for index, line in enumerate(lines):
            if variable.is_member_of:
                if line.lower().count("type") == 1 and line.lower().split()[-1] == variable.is_member_of:
                    condition1 = True
                condition2 = line.count("::")
                if condition1 and condition2:
                    m = re.search(pattern, line, re.I)
                    if m:
                        precision_specificator = m.group(1)
                        if precision_specificator.count("="):
                            precision_specificator = precision_specificator.split("=")[1]
                        try:
                            precision = lookup_table[precision_specificator.strip()]
                        except KeyError:
                            print("Error, you need to add %s to lookup table",precision_specificator)

                        variable.precision = precision
                        return True

                if line.lower().count("end type") and condition1:
                    raise AssertionError("Member %s not found in structure %s", variable, variable.is_member_of.name)
            else:
                if current_block is not None:
                    condition1 = variable.routine.name == current_block.name
                    condition2 = line.count("::")
                    condition3 = line.count(variable.name)
                    if condition1 and condition2:
                        m = re.search(pattern, line, re.I)
                        if m:
                            precision_specificator = m.group(1)
                            try:
                                precision = lookup_table[precision_specificator]
                            except KeyError:
                                print("Error, you probably need to add %s to lookup table", precision_specificator)

                            variable.precision = precision
                            return True
                        elif condition3:
                            cases += "%s -> %s\n" % (line.strip(), variable.name)
                current_block = in_which_block(line, current_block, vault=vault)
        # In case tha variable hasn't been found we check if it is a function with inline type declaration
        is_function = variable.name == variable.routine.name
        if not is_function:
            print("\r%s from routine %s in module %s not found!" \
                  % (variable.name, variable.routine.name, variable.routine.module.name))
            print("Instead we found:\n%s" % cases)
        else:
            pattern = r"function.*\b%s\b" % variable.name
            real_pattern = r"real *\(([^\)]*)\) *function.*\b%s\b" % variable.name
            for index, line in enumerate(lines):
                m = re.search(pattern, line, re.I)
                if m:
                    m2 = re.search(real_pattern, line, re.I)
                    if m2:
                        variable.precision = lookup_table[m2.group(1)]
                        return True
            print("\r%s's is a function. Type not found!" % variable.name)
            return False


if __name__ == "__main__":
    # Getting Command line options
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option("-v", "--vault", dest="vault", default=None, metavar="FILE")
    parser.add_option("-w", "--working-precision", dest="working_precision", default=23)
    parser.add_option("-o", "--output-vault", dest="output_vault", default="vault.pkl", metavar="FILE")
    parser.add_option("-s", "--input-source", dest="input_source",
                      default=None,metavar="FILE")

    (options, args) = parser.parse_args()
    vault_path = options.vault
    wp = options.working_precision
    output_vault = options.output_vault
    path_to_input_sources = options.input_source

    if path_to_input_sources is None and vault_path is None:
        print("\nOne of the arguments -s/--input-source or -v/--vault is mandatory\n")
        parser.print_help()
        exit(1)

    print("Creating a database from sources found in:\n   %s\n\n" % path_to_input_sources)

    # In case no vault has been provided it will create it from the sources in the input sources folder
    if vault_path is None:
        # Paths to source folders
        # Path_to_input_sources
        path_to_preprocessed_sources = "./tmp_processedsources/"
        path_to_output_sources = "./tmp_outputsources/"
        if not isdir(path_to_preprocessed_sources):
            mkdir(path_to_preprocessed_sources)
        if not isdir(path_to_output_sources):
            mkdir(path_to_output_sources)
        # List of files that should not be processed, as red from "list_of_files_to_keep_unmodified.txt"
        preprocess_blacklist = listdir(path_to_input_sources)

        # Format text + replace real declarations with type(rpe_var)
        preprocess_sources(path_to_input_sources, path_to_preprocessed_sources, blacklist=preprocess_blacklist)

        # Obtain source information
        vault = parse_sources(path_to_preprocessed_sources, save_vault=False)
    else:
        vault = load_vault(vault_path)

    # Lookup table with the significant bits value that correspond to each precision specifier
    lookup_table = {'WP': wp, 'default': wp, '8': 52, 'dp': 52, 'sp': 23, 'fbsp': 23, 'fbdp': 52, 'wp': wp, 'KIND=dp': 52,
                    "KIND=wp": wp,  'kind=fbsp': 23, "kind=sp": 23, "kind=wp": wp, "KIND=8": 23}

    # Searching again through the sources to find the type (single or double) of each variable in the vault
    print("Re-Finding variable types")

    real_types = ["single", "double"]
    real_variables = [v for v in vault.variables if v.type in real_types]
    length = len(real_variables)

    # Assigning id and precision
    type_to_precision = {"single": 23, "double":52}
    for index, var in enumerate(real_variables):
        var.id = index
        var.precision = type_to_precision[var.type]

    # Track dependencies and propagate it (will be modified with the graph approach)
    track_dependencies(path_to_preprocessed_sources, vault)
    propagate_dependencies(vault)

    # Save the vault with the variable precisions updated.
    # This workaround is needed to allow pickle to write the vault file
    import sys
    import pickle

    sys.setrecursionlimit(20000)
    pickle.dump(vault, open(output_vault, "wb"))

    # Deleting temporary folders
    if isdir(path_to_preprocessed_sources):
        rmtree(path_to_preprocessed_sources)
    if isdir(path_to_output_sources):
        rmtree(path_to_output_sources)
